
/*
 *
 *          Copyright (C) 1995, M. A. Sridhar
 *  
 *
 *     This software is Copyright M. A. Sridhar, 1995. You are free
 *     to copy, modify or distribute this software  as you see fit,
 *     and to use  it  for  any  purpose, provided   this copyright
 *     notice and the following   disclaimer are included  with all
 *     copies.
 *
 *                        DISCLAIMER
 *
 *     The author makes no warranties, either expressed or implied,
 *     with respect  to  this  software, its  quality, performance,
 *     merchantability, or fitness for any particular purpose. This
 *     software is distributed  AS IS.  The  user of this  software
 *     assumes all risks  as to its quality  and performance. In no
 *     event shall the author be liable for any direct, indirect or
 *     consequential damages, even if the  author has been  advised
 *     as to the possibility of such damages.
 *
 */



#if defined(__GNUC__)
#pragma implementation
#endif

#include "ui/cntroler.h"
#include "ui/evtbind.h"

#include "ui/support/x_yacl/window.h"
#include "ui/support/x_yacl/picture.h"

#include <X11/Xlib.h>
#include <X11/X.h>

typedef UI_EventBinding0 <UI_Picture> PictureBinding ;
template class UI_EventBinding0 <UI_Picture>         ;

UI_Picture::UI_Picture (UI_VisualObject* parent, const UI_Rectangle& r, 
                        char* bmp_bits)
{
    _parent    = parent                  ;
    _bmpbits   = bmp_bits                ;
    _bmprect   = r                       ;
    _fgChanged = false                   ;
    _bgChanged = false                   ;

    if (parent) {
        PictureBinding bind (this, &UI_Picture::_DrawPicture) ;
        parent -> AddEventDependent(Event_Paint,bind) ;
    }
}

void UI_Picture::SetForeground(const UI_Color& fg)
{
    _fgColor = fg ;
    _fgChanged = true ;
}

void UI_Picture::SetBackground(const UI_Color& bg)
{
    _bgColor = bg ;
    _bgChanged = true ;
}

bool UI_Picture::_DrawPicture()
{
    if (!_parent)
        return FALSE ;

    Display* dp = YACLApp() -> AppDisplay() ;
    int screenNo = DefaultScreen(dp) ;

    Window pw = _parent -> ViewHandle() ;
  
    if (!_bgChanged)
        _bgColor = _parent -> Background() ;
    if (!_fgChanged)
        _fgColor = _parent -> Foreground() ;

    Window root_ret ;
    int x_ret, y_ret ;
    unsigned int width_ret, height_ret, border_width_ret, depth_ret ;
    XGetGeometry(dp, pw, &root_ret, &x_ret, &y_ret, &width_ret, &height_ret,
                 &border_width_ret, &depth_ret) ;
    Pixmap stipple = 
        XCreatePixmapFromBitmapData(dp, pw, _bmpbits, _bmprect.Width(),
                                    _bmprect.Height(),_fgColor.XPixel(),
                                    _bgColor.XPixel(),depth_ret) ;
    GC gc ;
    XGCValues xgcv ;
    unsigned long gcMask ;
     
    gcMask = GCForeground | GCBackground ;
    xgcv.foreground = _fgColor.XPixel() ;
    xgcv.background = _bgColor.XPixel() ;
    
    gc = XCreateGC(dp,pw,gcMask,&xgcv) ;

    XCopyArea(dp, stipple, pw, gc, 0,0, 
              _bmprect.Width(), _bmprect.Height(), 
              _bmprect.Left(), _bmprect.Top()) ;

    XFreeGC(dp,gc) ;
    XFreePixmap(dp,stipple) ;
}

UI_Picture::~UI_Picture ()
{
}
