

/*
 *
 *          Copyright (C) 1997, M. A. Sridhar
 *  
 *
 *     This software is Copyright M. A. Sridhar, 1997. You are free
 *     to copy, modify or distribute this software  as you see fit,
 *     and to use  it  for  any  purpose, provided   this copyright
 *     notice and the following   disclaimer are included  with all
 *     copies.
 *
 *                        DISCLAIMER
 *
 *     The author makes no warranties, either expressed or implied,
 *     with respect  to  this  software, its  quality, performance,
 *     merchantability, or fitness for any particular purpose. This
 *     software is distributed  AS IS.  The  user of this  software
 *     assumes all risks  as to its quality  and performance. In no
 *     event shall the author be liable for any direct, indirect or
 *     consequential damages, even if the  author has been  advised
 *     as to the possibility of such damages.
 *
 */



#if defined(__GNUC__)
#pragma implementation
#endif


#include "ui/treevmdl.h"
#include "ui/treeview.h"

#include "base/treewalk.h"



typedef CL_ObjectPtrTree     UI_TreeViewModelBaseClass;

UI_TreeViewModel::UI_TreeViewModel ()
{
}


void UI_TreeViewModel::AddClient (UI_TreeView* client)
{
    _clientSet.Add (client);
}

void UI_TreeViewModel::RemoveClient (UI_TreeView* client)
{
    _clientSet.Remove (client);
}


CL_ObjectPtrTreeNode* UI_TreeViewModel::AddChild
    (CL_TreeNodeLabel child, CL_TreeNodeLabel parent, long i)
{
    CL_ObjectPtrTreeNode* node = UI_TreeViewModelBaseClass::AddChild
        (child, parent, i);
    if (node) {
        CL_ObjectSetIterator itr (_clientSet);
        while (itr.More()) {
            UI_TreeView* view = (UI_TreeView*) itr.Next();
            view->ModelHasAddedNode (node);
        }
    }
    return node;
}

CL_ObjectPtrTreeNode* UI_TreeViewModel::NewRoot
    (CL_TreeNodeLabel l)
{
    CL_ObjectPtrTreeNode* node = UI_TreeViewModelBaseClass::NewRoot (l);
    if (node) {
        CL_ObjectSetIterator itr (_clientSet);
        while (itr.More()) {
            UI_TreeView* view = (UI_TreeView*) itr.Next();
            view->ModelHasNewRoot ();
        }
    }
    return node;

}

UI_TreeViewModelBaseClass* UI_TreeViewModel::ExtractSubtree
    (CL_TreeNodeLabel label)
{
    CL_ObjectPtrTreeNode* node = Node (label);
    if (node) {
        CL_ObjectSetIterator itr (_clientSet);
        while (itr.More()) {
            UI_TreeView* view = (UI_TreeView*) itr.Next();
            view->ModelWillDestroySubtree (node);
        }
    }
    return UI_TreeViewModelBaseClass :: ExtractSubtree (label);
}


bool UI_TreeViewModel::DestroySubtree (CL_TreeNodeLabel x)
{
    CL_ObjectPtrTreeNode* node = Node (x);
    if (node && node != Root()) {
        CL_ObjectSetIterator itr (_clientSet);
        while (itr.More()) {
            UI_TreeView* view = (UI_TreeView*) itr.Next();
            view->ModelWillDestroySubtree (node);
        }
        UI_TreeViewModelBaseClass :: DestroySubtree (x);
        return TRUE;
    }
    return FALSE;
}


void UI_TreeViewModel::DestroyContents (CL_TreeNodeLabel x)
{
    CL_ObjectPtrTreeNode* node = Node (x);
    if (node) {
        CL_ObjectPtrTreePostWalker walker (node);
        while (walker.More()) {
            CL_ObjectPtrTreeNode* aNode = walker.Next();
            delete aNode->Content();
        }
        CL_ObjectSetIterator itr (_clientSet);
        while (itr.More()) {
            UI_TreeView* view = (UI_TreeView*) itr.Next();
            //   view->ModelHasDestroyedSubtree (x);
        }
    }
}


void UI_TreeViewModel::NodeContentChanged (CL_ObjectPtrTreeNode* node)
{
    CL_ObjectSetIterator itr (_clientSet);
    while (itr.More()) {
        UI_TreeView* view = (UI_TreeView*) itr.Next();
        view->ModelNodeContentChanged (node);
    }
}

